@@ -1,11 +1,11 @@
 use strict;
 use warnings;
-package JSON::Schema::Modern; # git description: v0.573-18-g6f7f97bf
+package JSON::Schema::Modern; # git description: v0.574-9-gade5d7e0
 # vim: set ts=8 sts=2 sw=2 tw=100 et :
 # ABSTRACT: Validate data against a schema
 # KEYWORDS: JSON Schema validator data validation structure specification
 
-our $VERSION = '0.574';
+our $VERSION = '0.575';
 
 use 5.020;  # for fc, unicode_strings features
 use Moo;
@@ -33,7 +33,7 @@ use Feature::Compat::Try;
 use JSON::Schema::Modern::Error;
 use JSON::Schema::Modern::Result;
 use JSON::Schema::Modern::Document;
-use JSON::Schema::Modern::Utilities qw(get_type canonical_uri E abort annotate_self);
+use JSON::Schema::Modern::Utilities qw(get_type canonical_uri E abort annotate_self jsonp is_type assert_uri);
 use namespace::clean;
 
 our @CARP_NOT = qw(
@@ -331,6 +331,8 @@ sub evaluate ($self, $data, $schema_reference, $config_override = {}) {
 
     abort($state, 'EXCEPTION: unable to find resource %s', $schema_reference)
       if not $schema_info;
+    abort($state, 'EXCEPTION: %s is not a schema', $schema_reference)
+      if not $schema_info->{document}->get_entity_at_location($schema_info->{document_path});
 
     $state = +{
       %$state,
@@ -464,10 +466,24 @@ sub _traverse_subschema ($self, $schema, $state) {
 
   my $valid = 1;
   my %unknown_keywords = map +($_ => undef), keys %$schema;
-  # we must check the array length on every iteration because some keywords can change it!
-  for (my $idx = 0; $idx <= $state->{vocabularies}->$#*; ++$idx) {
-    my $vocabulary = $state->{vocabularies}[$idx];
 
+  # First, we must determine the dialect to use. This is given to us in metaschema_uri
+  # and can also be indicated with the '$schema' keyword. We need to do this now, before iterating
+  # over vocabulary classes and keywords, because these can change depending on the dialect.
+  if (exists $schema->{'$schema'}) {
+    return if not $self->_parse_keyword_schema($state, $schema->{'$schema'});
+
+    # This is a bit of a chicken-and-egg situation. If we start off at draft2020-12, then all
+    # keywords are valid, so we inspect and process the $schema keyword; this switches us to draft7
+    # but now only the $ref keyword is respected and everything else should be ignored, so the
+    # $schema keyword never happened, so now we're back to draft2020-12 again, and...?!
+    # The only winning move is not to play.
+    return E($state, '$schema and $ref cannot be used together in older drafts')
+      if exists $schema->{'$ref'} and $state->{spec_version} eq 'draft7';
+  }
+
+  ALL_KEYWORDS:
+  foreach my $vocabulary ($state->{vocabularies}->@*) {
     # [ [ $keyword => $subref ], [ ... ] ]
     my $keyword_list = $vocabulary_cache->{$state->{spec_version}}{$vocabulary}{traverse} //= [
       map [ $_ => $vocabulary->can('_traverse_keyword_'.($_ =~ s/^\$//r)) ],
@@ -757,7 +773,7 @@ sub __all_metaschema_vocabulary_classes { values $_[0]->__metaschema_vocabulary_
 # retrieves metaschema info either from cache or by parsing the schema for vocabularies
 # throws a JSON::Schema::Modern::Result on error
 sub _get_metaschema_info ($self, $metaschema_uri, $for_canonical_uri) {
-  # check the cache
+  # check the cache. specification metaschemas are already populated.
   my $metaschema_info = $self->_get_metaschema_vocabulary_classes($metaschema_uri);
   return @$metaschema_info if $metaschema_info;
 
@@ -783,9 +799,96 @@ sub _get_metaschema_info ($self, $metaschema_uri, $for_canonical_uri) {
       $state->{errors}->@* ],
     exception => 1,
   ) if $state->{errors}->@*;
+
   return ($state->{spec_version}, $state->{vocabularies});
 }
 
+# we can't do this work in the context of looping over vocabularies and keywords, because this
+# keyword changes which vocabularies and keywords we're going to use. Additionally we may need to
+# fetch and parse the referenced schema to discover what vocabularies it defines.
+sub _parse_keyword_schema ($self, $state, $metaschema_uri) {
+  $state->{keyword} = '$schema';
+
+  return E($state, '$schema value is not a string') if not is_type('string', $metaschema_uri);
+  return if not assert_uri($state, { '$schema' => $metaschema_uri });
+
+  my ($spec_version, $vocabularies);
+
+  if (my $metaschema_info = $self->_get_metaschema_vocabulary_classes($metaschema_uri)) {
+    ($spec_version, $vocabularies) = @$metaschema_info;
+  }
+  else {
+    my $schema_info = $self->_fetch_from_uri($metaschema_uri);
+    return E($state, 'EXCEPTION: unable to find resource %s', $metaschema_uri) if not $schema_info;
+    # this cannot happen unless there are other entity types in the index
+    return E($state, 'EXCEPTION: bad reference to $schema %s: not a schema', $schema_info->{canonical_uri})
+      if $schema_info->{document}->get_entity_at_location($schema_info->{document_path}) ne 'schema';
+
+    if (not is_plain_hashref($schema_info->{schema})) {
+      ()= E($state, 'metaschemas must be objects');
+    }
+    else {
+      ($spec_version, $vocabularies) = $self->_fetch_vocabulary_data({ %$state,
+          keyword => '$vocabulary', initial_schema_uri => Mojo::URL->new($metaschema_uri),
+          traversed_schema_path => jsonp($state->{schema_path}, '$schema'),
+        }, $schema_info);
+    }
+  }
+
+  return E($state, '"%s" is not a valid metaschema', $metaschema_uri)
+    if not $vocabularies or not @$vocabularies;
+
+  $state->@{qw(spec_version vocabularies)} = ($spec_version, $vocabularies);
+  return 1;
+}
+
+# translate vocabulary URIs into classes, caching the results (if any)
+sub _fetch_vocabulary_data ($self, $state, $schema_info) {
+  if (not exists $schema_info->{schema}{'$vocabulary'}) {
+    # "If "$vocabulary" is absent, an implementation MAY determine behavior based on the meta-schema
+    # if it is recognized from the URI value of the referring schema's "$schema" keyword."
+    my $metaschema_uri = $self->METASCHEMA_URIS->{$schema_info->{specification_version}};
+    return $self->_get_metaschema_vocabulary_classes($metaschema_uri)->@*;
+  }
+
+  my $valid = 1;
+  $valid = E($state, '$vocabulary can only appear at the document root') if length $schema_info->{document_path};
+  $valid = E($state, 'metaschemas must have an $id') if not exists $schema_info->{schema}{'$id'};
+
+  return (undef, []) if not $valid;
+
+  my @vocabulary_classes;
+
+  foreach my $uri (sort keys $schema_info->{schema}{'$vocabulary'}->%*) {
+    my $class_info = $self->_get_vocabulary_class($uri);
+    $valid = E({ %$state, _schema_path_suffix => $uri }, '"%s" is not a known vocabulary', $uri), next
+      if $schema_info->{schema}{'$vocabulary'}{$uri} and not $class_info;
+
+    next if not $class_info;  # vocabulary is not known, but marked as false in the metaschema
+
+    my ($spec_version, $class) = @$class_info;
+    $valid = E({ %$state, _schema_path_suffix => $uri }, '"%s" uses %s, but the metaschema itself uses %s',
+        $uri, $spec_version, $schema_info->{specification_version}), next
+      if $spec_version ne $schema_info->{specification_version};
+
+    push @vocabulary_classes, $class;
+  }
+
+  @vocabulary_classes = sort {
+    $a->evaluation_order <=> $b->evaluation_order
+    || ($a->evaluation_order == 999 ? 0
+      : ($valid = E($state, '%s and %s have a conflicting evaluation_order', sort $a, $b)))
+  } @vocabulary_classes;
+
+  $valid = E($state, 'the first vocabulary (by evaluation_order) must be Core')
+    if ($vocabulary_classes[0]//'') ne 'JSON::Schema::Modern::Vocabulary::Core';
+
+  $self->_set_metaschema_vocabulary_classes($schema_info->{canonical_uri},
+    [ $schema_info->{specification_version}, \@vocabulary_classes ]) if $valid;
+
+  return ($schema_info->{specification_version}, $valid ? \@vocabulary_classes : []);
+}
+
 # used for determining a default '$schema' keyword where there is none
 use constant METASCHEMA_URIS => {
   'draft2020-12' => 'https://json-schema.org/draft/2020-12/schema',
@@ -912,7 +1015,7 @@ has _json_decoder => (
   default => sub { JSON::MaybeXS->new(allow_nonref => 1, canonical => 1, utf8 => 1, allow_bignum => 1, convert_blessed => 1) },
 );
 
-# since media types are case-insensitive, all type names must be foldcased on insertion.
+# since media types are case-insensitive, all type names must be casefolded on insertion.
 has _media_type => (
   is => 'bare',
   isa => my $media_type_type = Map[Str->where(q{$_ eq CORE::fc($_)}), CodeRef],
@@ -1019,7 +1122,7 @@ JSON::Schema::Modern - Validate data against a schema
 
 =head1 VERSION
 
-version 0.574
+version 0.575
 
 =head1 SYNOPSIS
 
@@ -1152,7 +1255,7 @@ the following keywords:
 
 =item *
 
-C<type> (where both C<string> and C<number> (and possibly C<integer>) are considered valid
+C<type> (where both C<string> and C<number> (and possibly C<integer>) are considered valid)
 
 =item *
 
@@ -1370,7 +1473,7 @@ C<metaschema_uri>: use the indicated URI as the metaschema
 
 You can pass a series of callback subs to this method corresponding to keywords, which is useful for
 extracting data from within schemas and skipping properties that may look like keywords but actually
-are not (for example C<{"const":{"$ref": "this is not actually a $ref"}}>). This feature is highly
+are not (for example C<{"const": {"$ref": "this is not actually a $ref"}}>). This feature is highly
 experimental and is highly likely to change in the future.
 
 For example, to find the resolved targets of all C<$ref> keywords in a schema document:
@@ -1414,6 +1517,9 @@ Adds support for a custom format. The data type that this format applies to must
 values of any other type will automatically be deemed to be valid, and will not be passed to the
 subref.
 
+Be careful to not mutate the type of the value while checking it -- for example, if it is a string,
+do not apply arithmetic operators to it -- or subsequent type checks on this value may fail.
+
 =head2 add_vocabulary
 
   $js->add_vocabulary('My::Custom::Vocabulary::Class');
@@ -1424,7 +1530,9 @@ L<"Meta-Schemas and Vocabularies"|https://json-schema.org/draft/2020-12/json-sch
 
 The class must compose the L<JSON::Schema::Modern::Vocabulary> role and implement the
 L<vocabulary|JSON::Schema::Modern::Vocabulary/vocabulary> and
-L<keywords|JSON::Schema::Modern::Vocabulary/keywords> methods.
+L<keywords|JSON::Schema::Modern::Vocabulary/keywords> methods, as well as
+C<< _traverse_keyword_<keyword name> >> methods for each keyword. C<< _eval_keyword_<keyword name> >>
+methods are optional; when not provided, evaluation will always return a true result.
 
 =head2 add_media_type
 
