@@ -4,7 +4,7 @@ package JSON::Schema::Modern::Vocabulary;
 # vim: set ts=8 sts=2 sw=2 tw=100 et :
 # ABSTRACT: Base role for JSON Schema vocabulary classes
 
-our $VERSION = '0.574';
+our $VERSION = '0.575';
 
 use 5.020;
 use Moo::Role;
@@ -74,6 +74,8 @@ sub eval ($class, $data, $schema, $state) {
 sub eval_subschema_at_uri ($class, $data, $schema, $state, $uri) {
   my $schema_info = $state->{evaluator}->_fetch_from_uri($uri);
   abort($state, 'EXCEPTION: unable to find resource %s', $uri) if not $schema_info;
+  abort($state, 'EXCEPTION: bad reference to %s: not a schema', $schema_info->{canonical_uri})
+    if $schema_info->{document}->get_entity_at_location($schema_info->{document_path}) ne 'schema';
 
   my $vocabularies = $schema_info->{vocabularies};
   if ($state->{validate_formats}) {
@@ -114,7 +116,7 @@ JSON::Schema::Modern::Vocabulary - Base role for JSON Schema vocabulary classes
 
 =head1 VERSION
 
-version 0.574
+version 0.575
 
 =head1 SYNOPSIS
 
