@@ -4,7 +4,7 @@ package JSON::Schema::Modern::Vocabulary::FormatAssertion;
 # vim: set ts=8 sts=2 sw=2 tw=100 et :
 # ABSTRACT: Implementation of the JSON Schema Format-Assertion vocabulary
 
-our $VERSION = '0.574';
+our $VERSION = '0.575';
 
 use 5.020;
 use Moo;
@@ -143,7 +143,7 @@ sub keywords {
     'relative-json-pointer' => sub { $_[0] =~ m{^(?:0|[1-9][0-9]*)(?:#$|$|/)} && $_[0] !~ m{~(?![01])} },
     regex => sub {
       local $SIG{__WARN__} = sub { die @_ };
-      eval { qr/$_[0]/; 1 ? 1 : 0 };
+      eval { qr/$_[0]/; 1 };
     },
 
     'iri-reference' => sub { 1 },
@@ -222,8 +222,8 @@ sub _eval_keyword_format ($class, $data, $schema, $state) {
       : undef;
 
   A($state, $schema->{format});
-  return E($state, 'not a%s %s', $schema->{format} =~ /^[aeio]/ ? 'n' : '', $schema->{format})
-    if $spec and is_type($spec->{type}, $data) and not $spec->{sub}->($data);
+  return E($state, 'not a valid %s', $schema->{format}) if $spec and is_type($spec->{type}, $data)
+    and not $spec->{sub}->($data);
 
   return 1;
 }
@@ -242,7 +242,7 @@ JSON::Schema::Modern::Vocabulary::FormatAssertion - Implementation of the JSON S
 
 =head1 VERSION
 
-version 0.574
+version 0.575
 
 =head1 DESCRIPTION
 
